==> cirq-0.4.0.dev4220181119182308-py2-none-any.whl: cirq/optimizers/convert_to_cz_and_single_gates.py <==
# coding=utf-8
# Copyright 2018 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from typing import Optional

from cirq import ops, decompositions, protocols
from cirq.circuits.circuit import Circuit
from cirq.circuits.optimization_pass import (
    PointOptimizationSummary,
    PointOptimizer,
)


class ConvertToCzAndSingleGates(PointOptimizer):
    u"""Attempts to convert strange multi-qubit gates into CZ and single qubit
    gates.

    First, checks if the operation has a unitary effect. If so, and the gate is
        a 1-qubit or 2-qubit gate, then performs circuit synthesis of the
        operation.

    Second, attempts to `cirq.decompose` to the operation.

    Third, if ignore_failures is set, gives up and returns the gate unchanged.
        Otherwise raises a TypeError.
    """

    def __init__(self,
                 ignore_failures = False,
                 allow_partial_czs = False):
        u"""
        Args:
            ignore_failures: If set, gates that fail to convert are forwarded
                unchanged. If not set, conversion failures raise a TypeError.
            allow_partial_czs: If set, the decomposition is permitted to use
                gates of the form `cirq.CZ**t`, instead of only `cirq.CZ`.
        """
        super(ConvertToCzAndSingleGates, self).__init__()
        self.ignore_failures = ignore_failures
        self.allow_partial_czs = allow_partial_czs

    def _keep(self, op):
        # Check if this is a CZ
        # Only keep partial CZ gates if allow_partial_czs
        if (isinstance(op, ops.GateOperation)
            and isinstance(op.gate, ops.CZPowGate)
            and (self.allow_partial_czs or op.gate.exponent == 1)):
            return True

        # Measurement?
        if ops.MeasurementGate.is_measurement(op):
            return True

        # SingleQubit known matrix
        if len(op.qubits) == 1 and protocols.has_unitary(op):
            return True
        return False

    def _decompose_two_qubit_unitaries(self, op):
        # Known matrix?
        if len(op.qubits) == 2:
            mat = protocols.unitary(op, None)
            if mat is not None:
                return decompositions.two_qubit_matrix_to_operations(
                    op.qubits[0],
                    op.qubits[1],
                    mat,
                    allow_partial_czs=self.allow_partial_czs)
        return NotImplemented

    def _on_stuck_raise(self, op):
        raise TypeError(u"Don't know how to work with {!r}. "
                        u"It isn't composite or an operation with a "
                        u"known unitary effect on 1 or 2 qubits.".format(op))

    def optimization_at(self, circuit, index, op
                        ):
        converted = protocols.decompose(
            op,
            intercepting_decomposer=self._decompose_two_qubit_unitaries,
            keep=self._keep,
            on_stuck_raise=(None
                            if self.ignore_failures
                            else self._on_stuck_raise))
        if converted == [op]:
            return None

        return PointOptimizationSummary(
            clear_span=1,
            new_operations=converted,
            clear_qubits=op.qubits)

==> cirq-0.5.0-py2-none-any.whl: cirq/optimizers/convert_to_cz_and_single_gates.py <==
# coding=utf-8
# Copyright 2018 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from typing import Optional

from cirq import circuits, ops, protocols, value
from cirq.optimizers import two_qubit_decompositions


class ConvertToCzAndSingleGates(circuits.PointOptimizer):
    u"""Attempts to convert strange multi-qubit gates into CZ and single qubit
    gates.

    First, checks if the operation has a unitary effect. If so, and the gate is
        a 1-qubit or 2-qubit gate, then performs circuit synthesis of the
        operation.

    Second, attempts to `cirq.decompose` to the operation.

    Third, if ignore_failures is set, gives up and returns the gate unchanged.
        Otherwise raises a TypeError.
    """

    def __init__(self,
                 ignore_failures = False,
                 allow_partial_czs = False):
        u"""
        Args:
            ignore_failures: If set, gates that fail to convert are forwarded
                unchanged. If not set, conversion failures raise a TypeError.
            allow_partial_czs: If set, the decomposition is permitted to use
                gates of the form `cirq.CZ**t`, instead of only `cirq.CZ`.
        """
        super(ConvertToCzAndSingleGates, self).__init__()
        self.ignore_failures = ignore_failures
        self.allow_partial_czs = allow_partial_czs

    def _keep(self, op):
        # Check if this is a CZ
        # Only keep partial CZ gates if allow_partial_czs
        cz_gate = ops.op_gate_of_type(op, ops.CZPowGate)
        if (cz_gate and (
                self.allow_partial_czs or value.canonicalize_half_turns(
                cz_gate.exponent) == 1)):
            return True

        # Measurement in Z basis?
        if ops.op_gate_of_type(op, ops.MeasurementGate):
            return True

        # SingleQubit known matrix
        if len(op.qubits) == 1 and protocols.has_unitary(op):
            return True
        return False

    def _decompose_two_qubit_unitaries(self, op):
        # Known matrix?
        if len(op.qubits) == 2:
            mat = protocols.unitary(op, None)
            if mat is not None:
                return two_qubit_decompositions.two_qubit_matrix_to_operations(
                    op.qubits[0],
                    op.qubits[1],
                    mat,
                    allow_partial_czs=self.allow_partial_czs)
        return NotImplemented

    def _on_stuck_raise(self, op):
        raise TypeError(u"Don't know how to work with {!r}. "
                        u"It isn't composite or an operation with a "
                        u"known unitary effect on 1 or 2 qubits.".format(op))

    def optimization_at(self,
                        circuit,
                        index,
                        op
                        ):
        converted = protocols.decompose(
            op,
            intercepting_decomposer=self._decompose_two_qubit_unitaries,
            keep=self._keep,
            on_stuck_raise=(None
                            if self.ignore_failures
                            else self._on_stuck_raise))
        if converted == [op]:
            return None

        return circuits.PointOptimizationSummary(
            clear_span=1,
            new_operations=converted,
            clear_qubits=op.qubits)
